# Imports from python.
from functools import reduce
import uuid


# Imports from Django.
from django.db import models
from django.utils import timezone


# Imports from other dependencies.
from uuslug import uuslug


# Imports from civic_utils.
from civic_utils.models.managers import BaseNaturalKeyManager


class UniqueIdentifierMixin(models.Model):
    uid_base_field = "slug"
    uid_prefix = ""

    uid = models.CharField(max_length=500, editable=False, blank=True)

    class Meta:
        abstract = True

    def get_uid_base_field(self):
        return getattr(self, self.uid_base_field)

    @classmethod
    def get_uid_prefix(self):
        return self.uid_prefix

    def get_uid_suffix(self):
        return self.get_uid_base_field()

    def generate_unique_identifier(self, always_overwrite_uid=False):
        """Generate slug and (if needed) UID field values."""
        generate_uid = always_overwrite_uid

        if not always_overwrite_uid and not self.uid:
            generate_uid = True

        if generate_uid:
            self.uid = "{}:{}".format(
                self.get_uid_prefix(), self.get_uid_suffix()
            )


class CommonIdentifiersMixin(UniqueIdentifierMixin):
    uid_base_field = "slug"

    uid = models.CharField(max_length=500, editable=False, blank=True)

    slug = models.SlugField(
        blank=True, max_length=255, unique=True, editable=False
    )

    class Meta:
        abstract = True

    def generate_common_identifiers(
        self, always_overwrite_slug=True, always_overwrite_uid=False
    ):
        """Generate slug and (if needed) UID field values."""
        generate_slug = always_overwrite_slug

        if not always_overwrite_slug and not self.slug:
            generate_slug = True

        if generate_slug:
            self.slug = uuslug(
                getattr(self, self.get_uid_base_field()),
                instance=self,
                max_length=100,
                separator="-",
                start_no=2,
            )

        self.generate_unique_identifier(always_overwrite_uid)


class NaturalKeyMixin(models.Model):
    """A mixin to add default natural-key behaviors to Civic models.

    Borrows somewhat heavily from the following project:
    https://github.com/wq/django-natural-keys/
    """

    natural_key_fields = []

    objects = BaseNaturalKeyManager()

    class Meta:
        abstract = True

    def get_natural_key_definition(cls):
        """Override for self.natural_key_fields list, if needed."""
        return cls.natural_key_fields

    @classmethod
    def get_natural_key_info(cls):
        """
        Derive natural key from first unique_together definition, noting
        which fields are related objects vs. regular fields.
        """
        fields = cls.get_natural_key_definition()
        info = []
        for name in fields:
            field = cls._meta.get_field(name)
            rel_to = None
            if hasattr(field, "rel"):
                rel_to = field.rel.to if field.rel else None
            elif hasattr(field, "remote_field"):
                if field.remote_field:
                    rel_to = field.remote_field.model
                else:
                    rel_to = None
            info.append((name, rel_to))
        return info

    @classmethod
    def get_natural_key_fields(cls):
        """
        Determine actual natural key field list, incorporating the natural keys
        of related objects as needed.
        """
        natural_key = []
        for name, rel_to in cls.get_natural_key_info():
            if not rel_to:
                natural_key.append(name)
            elif issubclass(rel_to, NaturalKeyMixin):
                nested_key = rel_to.get_natural_key_fields()
                natural_key.extend(
                    [f"{name}__{nested_name}" for nested_name in nested_key]
                )
            else:
                natural_key.append(f"{name}__pk")
        return natural_key

    def natural_key(self):
        key_field_values = [
            reduce(getattr, name.split("__"), self)
            for name in self.get_natural_key_fields()
        ]

        return key_field_values


class UUIDMixin(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True



# flake8: noqa

# Imports from Django.
from django.db import models
from django.core import serializers as django_serializers


# Imports from civic_utils.
from civic_utils.models.managers import CivicBaseManager
from civic_utils.models.mixins import NaturalKeyMixin
from civic_utils.utils.importers import import_class
from civic_utils.utils.rest_formats import REST_FORMAT_PARSERS
from civic_utils.utils.rest_formats import REST_FORMAT_RENDERERS


class CivicBaseModel(NaturalKeyMixin, models.Model):
    created = models.DateTimeField(editable=False)
    updated = models.DateTimeField(editable=False)

    objects = CivicBaseManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self.id:
            self.created = timezone.now()
        self.updated = timezone.now()

        return super(TimeTrackedMixin, self).save(*args, **kwargs)

    @classmethod
    def get_default_serializer(self, format="json"):
        serializer_path = getattr(self, "default_serializer", None)

        if serializer_path and format in REST_FORMAT_RENDERERS.keys():
            serializer_klass = import_class(serializer_path)
            renderer_klass = REST_FORMAT_RENDERERS[format]

            return serializer_klass, renderer_klass

        return (django_serializers.get_serializer(format), None)

    @classmethod
    def deserialize(self, format, data_stream):
        parser_klass = REST_FORMAT_PARSERS[format]
        raw_data = parser_klass().parse(data_stream)

        serializer_path = getattr(self, "default_serializer", None)

        if serializer_path:
            serializer_klass = import_class(serializer_path)
            return serializer_klass(many=True).deserialize(raw_data)

        return list(django_serializers.python.Deserializer(raw_data))



# Imports from other dependencies.
import boto3


# Imports from civic_utils.
from civic_utils.conf import settings


def get_s3_session():
    return boto3.Session(
        aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
        aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
    )


def get_bucket(bucket_name):
    session = get_s3_session()
    s3 = session.resource("s3")
    return s3.Bucket(bucket_name)



# Imports from civic_utils.
from civic_utils.utils.rest_formats import REST_FORMAT_PARSERS
from civic_utils.utils.rest_formats import REST_FORMAT_RENDERERS


DEFAULT_FIXTURE_ACL = 'bucket-owner-full-control'

TEMP_DIR = 'tmp'

ALLOWED_SERIALIZER_FORMATS = list(REST_FORMAT_PARSERS.keys())

FORMAT_EXTENSION_MAP = {
    'json': 'json'
}

FORMAT_MIME_MAP = dict(
    (format_str, renderer_klass.media_type)
    for format_str, renderer_klass in REST_FORMAT_RENDERERS.items()
)



# Imports from Django.
from django.db import models


# Imports from other dependencies.
from rest_framework.serializers import ModelSerializer


class CivicBaseQuerySet(models.QuerySet):
    """"""

    def serialize(self, format="json", **export_kwargs):
        # dict(
        #     indent
        #     stream
        #     progress_output
        #     object_count
        # )

        (serializer_klass, renderer_klass) = self.model.get_default_serializer(
            format
        )

        if issubclass(serializer_klass, ModelSerializer) and renderer_klass:
            return serializer_klass(self, many=True).serialize_to_cli(
                renderer=renderer_klass, **export_kwargs
            )

        return serializer_klass().serialize(
            self,
            **dict(
                use_natural_foreign_keys=True,
                use_natural_primary_keys=True,
                **export_kwargs,
                # stream=stream or self.stdout,
                # progress_output=progress_output,
                # object_count=object_count,
            ),
        )


class NaturalKeyQuerySetMixin(object):
    """A mixin to add default natural-key behaviors to Civic QuerySets.

    Borrows somewhat heavily from the following project:
    https://github.com/wq/django-natural-keys/
    """

    def natural_key_kwargs(self, *args):
        """Turn args >> kwargs by merging with the natural key fields.

        Takes a list of args, which gets zipped together with the
        list generated by the model's get_natural_key_fields() method.

        Also checks to make sure the right number of args are submitted
        (otherwise Django doesn't enforce this, which can lead to only
        partial matches being evaluated and returned).
        """
        natural_key_fields = self.model.get_natural_key_fields()

        if len(args) != len(natural_key_fields):
            raise TypeError(
                "Mismatched number of args and natural-key fields: "
                f"expected {len(natural_key_fields)} argument(s)."
            )

        return dict(zip(natural_key_fields, args))


class BaseNaturalKeyQuerySet(NaturalKeyQuerySetMixin, CivicBaseQuerySet):
    pass



# Imports from Django.
from django.db import models


# Imports from civic_utils.
from civic_utils.models.helpers import extract_nested_key
from civic_utils.models.querysets import BaseNaturalKeyQuerySet
from civic_utils.models.querysets import CivicBaseQuerySet


class CivicBaseManager(models.Manager):
    """"""

    def get_queryset(self):
        return CivicBaseQuerySet(self.model, using=self._db)


class NaturalKeyManagerMixin(object):
    """A mixin to add default natural-key behaviors to Civic managers.

    Borrows somewhat heavily from the following project:
    https://github.com/wq/django-natural-keys/
    """

    def get_queryset(self):
        return BaseNaturalKeyQuerySet(self.model, using=self._db)

    def natural_key_kwargs(self, *args):
        """Turn args >> kwargs by merging with the natural key fields.

        Falls through to the method with the same name on the QuerySet.
        """
        return self.get_queryset().natural_key_kwargs(*args)

    def get_by_natural_key(self, *args):
        """Return the object corresponding to the provided natural key.

        (Generic implementation of the standard Django function).
        """
        natural_kwargs = self.natural_key_kwargs(*args)

        # Since kwargs already has __ lookups in it, we could just do this:
        # return self.get(**natural_kwargs)

        # But, we should call each related model's get_by_natural_key() method,
        # in case it's been overridden.
        for name, rel_to in self.model.get_natural_key_info():
            if not rel_to:
                continue

            # Extract natural key for related object
            nested_key = extract_nested_key(natural_kwargs, rel_to, name)
            if nested_key:
                # Update natural_kwargs with related object
                try:
                    natural_kwargs[name] = rel_to.objects.get_by_natural_key(
                        *nested_key
                    )
                except rel_to.DoesNotExist:
                    # If related object doesn't exist, assume the model we're
                    # querying for doesn't exist, either.
                    raise self.model.DoesNotExist()
            else:
                natural_kwargs[name] = None

        return self.get(**natural_kwargs)


class BaseNaturalKeyManager(NaturalKeyManagerMixin, models.Manager):
    pass



# Imports from python.
import os
import time
import warnings


# Imports from Django.
from django.apps import apps
from django.core import serializers
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.core.management.utils import parse_apps_and_model_labels
from django.db import DEFAULT_DB_ALIAS
from django.db import router


# Imports from other dependencies.
import crayons
from progress.bar import ChargingBar


# Imports from civic_utils.
from civic_utils.conf import settings
from civic_utils.management.helpers import ALLOWED_SERIALIZER_FORMATS
from civic_utils.management.helpers import DEFAULT_FIXTURE_ACL
from civic_utils.management.helpers import FORMAT_EXTENSION_MAP
from civic_utils.management.helpers import FORMAT_MIME_MAP
from civic_utils.management.helpers import TEMP_DIR
from civic_utils.models import CivicBaseModel
from civic_utils.utils.aws import get_bucket


S3_FIXTURE_BUCKET = settings.AWS_S3_BUCKET

S3_FIXTURE_ROOT_PATH = settings.FIXTURE_ROOT_PATH


class ProxyModelWarning(Warning):
    pass


def get_civic_models_for_app(app_config_obj):
    return [
        model for model in app_config_obj.get_models()
        if issubclass(model, CivicBaseModel)
    ]


class Command(BaseCommand):
    help = (
        "Export selected contents of the database as fixtures uploaded to S3. "
        "Fixtures will be stored in the user-specified format (using each "
        "model's default manager unless --all is specified)."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            'args', metavar='app_label[.ModelName]', nargs='*',
            help=(
                'Restricts dumped data to the specified app_label or '
                'app_label.ModelName.'
            ),
        )
        parser.add_argument(
            '--format', default='json',
            help='Specifies the output serialization format for fixtures.',
        )
        parser.add_argument(
            '--indent', type=int,
            help=(
                'Specifies the indent level to use when '
                'pretty-printing output.'
            ),
        )
        parser.add_argument(
            '--database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a specific database to dump fixtures from. '
                 'Defaults to the "default" database.',
        )
        parser.add_argument(
            '-e', '--exclude', action='append', default=[],
            help=(
                'An app_label or app_label.ModelName to exclude '
                '(use multiple --exclude to exclude multiple apps/models).'
            ),
        )
        parser.add_argument(
            '-l', '--local', action='store_true', dest='store_locally',
            help='Store fixtures locally, rather than uploading them to S3.',
        )
        parser.add_argument(
            '-a', '--all', action='store_true', dest='use_base_manager',
            help=(
                "Use Django's base manager to dump all models stored in the "
                "database, including those that would otherwise be filtered "
                "or modified by a custom manager."
            ),
        )
        parser.add_argument(
            '-f', '--filename',
            default='export',
            dest='output_filename',
            help='Specifies file to which the output is written.'
        )

    def handle(self, *app_labels, **options):
        format = options['format']
        indent = options['indent']
        using = options['database']
        excludes = options['exclude']
        show_traceback = options['traceback']
        store_locally = options['store_locally']
        use_base_manager = options['use_base_manager']
        output_filename = options['output_filename']

        excluded_models, excluded_apps = parse_apps_and_model_labels(excludes)

        if not app_labels:
            all_apps_with_models = [
                app_config for app_config in apps.get_app_configs()
                if app_config.models_module is not None
                and app_config not in excluded_apps
            ]

            # This would be an excellent candidate for assignment expressions
            # (introduced via PEP 572 in Python 3.8) once we've upgraded.
            app_list = dict(
                (app_config, get_civic_models_for_app(app_config))
                for app_config in all_apps_with_models
                if get_civic_models_for_app(app_config)
            )
        else:
            app_list = {}
            for label in app_labels:
                try:
                    app_label, model_label = label.split('.')
                    try:
                        app_config = apps.get_app_config(app_label)
                    except LookupError as e:
                        raise CommandError(str(e))

                    if app_config.models_module is None:
                        continue
                    elif app_config in excluded_apps:
                        continue

                    try:
                        model = app_config.get_model(model_label)
                    except LookupError:
                        raise CommandError(
                            "Unknown model: %s.%s" % (app_label, model_label)
                        )

                    app_list_value = app_list.setdefault(app_config, [])

                    # We may have previously seen an "all-models" request for
                    # this app (no model qualifier was given). In this case
                    # there is no need adding specific models to the list.
                    if app_list_value is not None:
                        if all([
                            model not in app_list_value,
                            issubclass(model, CivicBaseModel)
                        ]):
                            app_list_value.append(model)

                except ValueError:
                    # This is just an app - no model qualifier
                    app_label = label
                    try:
                        app_config = apps.get_app_config(app_label)
                    except LookupError as e:
                        raise CommandError(str(e))

                    if app_config.models_module is None:
                        continue
                    elif app_config in excluded_apps:
                        continue

                    app_list[app_config] = get_civic_models_for_app(app_config)

        # Check that the serialization format exists; this is a shortcut to
        # avoid collating all the objects and _then_ failing.
        if format not in ALLOWED_SERIALIZER_FORMATS:
            try:
                serializers.get_serializer(format)
            except serializers.SerializerDoesNotExist:
                pass

            raise CommandError("Unknown serialization format: %s" % format)

        output_extension = FORMAT_EXTENSION_MAP[format]

        models_to_export = serializers.sort_dependencies(app_list.items())

        def get_objects(count_only=False):
            """
            Collate the objects to be serialized. If count_only is True, just
            count the number of objects to be serialized.
            """
            for model in models_to_export:
                if model in excluded_models:
                    continue
                if all([
                    model._meta.proxy,
                    model._meta.proxy_for_model not in models_to_export,
                ]):
                    warnings.warn(
                        "%s is a proxy model and won't be serialized." % model._meta.label,
                        category=ProxyModelWarning,
                    )
                if not model._meta.proxy and router.allow_migrate_model(using, model):
                    if use_base_manager:
                        objects = model._base_manager
                    else:
                        objects = model._default_manager

                    queryset = objects.using(using).order_by(model._meta.pk.name)

                    if count_only:
                        yield queryset.order_by().count()
                    else:
                        yield queryset

        model_counts = [count for count in get_objects(count_only=True)]

        self.stdout.write('')
        self.stdout.write(crayons.magenta('Civic Data Exporter', bold=True))
        self.stdout.write('')

        directory_label = 'Local' if store_locally else 'Temporary local'
        self.stdout.write(f"• {directory_label} directory: './{TEMP_DIR}/'")

        if store_locally:
            self.stdout.write(f"• No S3 exports (storing locally instead)")
        else:
            self.stdout.write(f"• S3 bucket: '{S3_FIXTURE_BUCKET}'")
            self.stdout.write(f"• S3 path: './{S3_FIXTURE_ROOT_PATH}/'")

        self.stdout.write('')
        self.stdout.write('')
        self.stdout.write((
            "📝  The following types and counts of models will be exported "
            f"({sum(model_counts)} total model instances):"
        ))
        self.stdout.write('')
        self.stdout.write('')

        counts_by_model_klass = dict(zip(models_to_export, model_counts))

        export_destinations = {
            model_config._meta.db_table: os.path.join(
                model_config._meta.model_name,
                f"{output_filename}.{output_extension}"
            )
            for model_config, instance_count in counts_by_model_klass.items()
        }

        for model_config, instance_count in counts_by_model_klass.items():
            model_file = export_destinations[model_config._meta.db_table]

            self.stdout.write(
                crayons.yellow(f"[{model_config._meta.label}]", bold=True)
            )
            self.stdout.write('')
            self.stdout.write((
                f"-   {instance_count} "
                f"instance{'s' if instance_count != 1 else ''}"
                f"{'; blank file only' if instance_count == 0 else ''}"
            ))
            self.stdout.write('-   ➡ (prefix)/{}'.format(
                crayons.cyan(f"{model_file}", bold=True),
            ))
            self.stdout.write('')
            self.stdout.write('')

        self.stdout.write(
            '----------------------------------'
            '----------------------------------'
        )
        self.stdout.write('')
        self.stdout.write('')

        try:
            progress_output = None
            object_count = 0

            # If dumpdata is outputting to stdout, there is no way to display progress
            if all([
                output_filename,
                self.stdout.isatty(),
                options['verbosity'] > 0
            ]):
                progress_output = self.stdout
                object_count = sum(model_counts)
                self.stdout.write('💾  {}'.format(
                    crayons.blue(
                        f'Exporting model instances to ./{TEMP_DIR}/:',
                        bold=True
                    )
                ))
                self.stdout.write('')

                progress_bar = ChargingBar(
                    '%(percent)d%%',
                    max=object_count if object_count > 0 else 1,
                    suffix=(
                        '[%(finished_models)d / '
                        f'{len(models_to_export)} instances done; %(eta)ds]'
                    ),
                    finished_models=0,
                )

            exported_files = []

            for model_queryset in get_objects():
                raw_destination = export_destinations[
                    model_queryset.model._meta.db_table
                ]
                full_destination = os.path.join(TEMP_DIR, raw_destination)

                os.makedirs(os.path.dirname(full_destination), exist_ok=True)

                per_model_output_file = full_destination

                with open(full_destination, 'w') as model_output:
                    model_queryset.serialize(
                        stream=model_output,
                        indent=indent,
                    )

                models_added = model_queryset.order_by().count()

                progress_bar.finished_models += 1

                progress_bar.next(n=models_added if models_added else 1)

                exported_files.append(raw_destination)

            progress_bar.finish()
            self.stdout.write('')
            self.stdout.write('')

        except Exception as e:
            if show_traceback:
                raise
            raise CommandError("Unable to serialize database: %s" % e)

        formatted_export_destination = f"'./{TEMP_DIR}/'"

        if not store_locally:
            exported_file_count = len(exported_files)

            self.stdout.write('🌎  {}'.format(
                crayons.blue('Uploading data files to S3:', bold=True)
            ))
            self.stdout.write('')

            upload_progress_bar = ChargingBar(
                '%(percent)d%%',
                max=exported_file_count if exported_file_count > 0 else 1,
                suffix='[%(index)d / %(max)d uploaded; %(eta)ds]',
            )

            formatted_export_destination = (
                f"'./{S3_FIXTURE_ROOT_PATH}/' in S3 bucket "
                f"'{S3_FIXTURE_BUCKET}'"
            )

            paths_to_remove = []
            for export_path in exported_files:
                full_export_path = os.path.join(TEMP_DIR, export_path)
                paths_to_remove.append(full_export_path)

                with open(full_export_path, 'r') as input_file:
                    bucket = get_bucket(S3_FIXTURE_BUCKET)
                    bucket.put_object(
                        ACL=DEFAULT_FIXTURE_ACL,
                        Body=input_file.read(),
                        ContentType=FORMAT_MIME_MAP[format],
                        Key=os.path.join(S3_FIXTURE_ROOT_PATH, export_path)
                    )
                upload_progress_bar.next()

            upload_progress_bar.finish()
            self.stdout.write('')
            self.stdout.write('')

            self.stdout.write("🗑️  {}".format(
                crayons.blue(f"Removing files from ./{TEMP_DIR}/:", bold=True)
            ))
            self.stdout.write("")

            temp_clearing_progress_bar = ChargingBar(
                '%(percent)d%%',
                max=len(paths_to_remove) if len(paths_to_remove) > 0 else 1,
                suffix='[%(index)d / %(max)d deleted; %(eta)ds]',
            )

            for local_fixture_path in paths_to_remove:
                time.sleep(0.5)
                os.remove(local_fixture_path)
                temp_clearing_progress_bar.next()

            temp_clearing_progress_bar.finish()
            self.stdout.write('')
            self.stdout.write('')

        self.stdout.write('🍻  {}'.format(
            crayons.green('Done!', bold=True)
        ))
        self.stdout.write('')
        self.stdout.write(crayons.green(
            f" Successfully exported {sum(model_counts)} model instances to "
            f"{formatted_export_destination}."
        ))
        self.stdout.write('')
        self.stdout.write('')



"""
Use this file to configure pluggable app settings and resolve defaults
with any overrides set in project settings.
"""

# Imports from Django.
from django.conf import settings as project_settings


class Settings:
    pass


Settings.AWS_ACCESS_KEY_ID = getattr(
    project_settings, "CIVIC_UTILS_AWS_ACCESS_KEY_ID", None
)

Settings.AWS_SECRET_ACCESS_KEY = getattr(
    project_settings, "CIVIC_UTILS_AWS_SECRET_ACCESS_KEY", None
)

Settings.AWS_S3_BUCKET = getattr(
    project_settings, "CIVIC_UTILS_AWS_S3_BUCKET", None
)

Settings.FIXTURE_ROOT_PATH = getattr(
    project_settings, "CIVIC_UTILS_FIXTURE_ROOT", None
)

settings = Settings



def extract_nested_key(key, cls, prefix=""):
    nested_key = cls.get_natural_key_fields()
    local_fields = {field.name: field for field in cls._meta.local_fields}
    values = []
    has_val = False
    if prefix:
        prefix += "__"
    for nname in nested_key:
        val = key.pop(prefix + nname, None)
        if val is None and nname in local_fields:
            if type(local_fields[nname]).__name__ == "DateTimeField":
                date = key.pop(nname + "_date", None)
                time = key.pop(nname + "_time", None)
                if date and time:
                    val = "%s %s" % (date, time)

        if val is not None:
            has_val = True
        values.append(val)
    if has_val:
        return values
    else:
        return None



# Imports from civic_utils.
from civic_utils.models.base import CivicBaseModel
from civic_utils.models.managers import BaseNaturalKeyManager
from civic_utils.models.mixins import CommonIdentifiersMixin
from civic_utils.models.mixins import NaturalKeyMixin
from civic_utils.models.mixins import UniqueIdentifierMixin
from civic_utils.models.mixins import UUIDMixin
from civic_utils.models.querysets import BaseNaturalKeyQuerySet


__all__ = [
    "BaseNaturalKeyManager",
    "BaseNaturalKeyQuerySet",
    "CivicBaseModel",
    "CommonIdentifiersMixin",
    "NaturalKeyMixin" "UniqueIdentifierMixin",
    "UUIDMixin",
]



# Imports from python.
from importlib import import_module


def import_class(val):
    """Import a class from a string module path.

    Pattern borrowed from Django REST Framework.
    See rest_framework/settings.py#L170-L182
    """
    try:
        parts = val.split(".")
        module_path, class_name = ".".join(parts[:-1]), parts[-1]
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import class '{}'. {}: {}.".format(
            val, e.__class__.__name__, e
        )
        raise ImportError(msg)



# Imports from other dependencies.
from rest_framework.settings import api_settings


REST_FORMAT_PARSERS = {
    parser_klass.renderer_class.format: parser_klass
    for parser_klass in api_settings.DEFAULT_PARSER_CLASSES
    if hasattr(parser_klass, "renderer_class")
}

REST_FORMAT_RENDERERS = {
    format_str: parser_klass.renderer_class
    for format_str, parser_klass in REST_FORMAT_PARSERS.items()
}



# Imports from python.
import os
from setuptools import find_packages
from setuptools import setup


# Imports from civic_utils.
from civic_utils import __version__


REPO_URL = "https://github.com/The-Politico/politico-civic-utils/"

PYPI_VERSION = ".".join(str(v) for v in __version__)

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()


setup(
    name="politico-civic-utils",
    version=PYPI_VERSION,
    packages=find_packages(exclude=["docs", "tests", "example"]),
    include_package_data=True,
    license="MIT",
    description="Shared utilities for POLITICO Civic apps",
    long_description=README,
    long_description_content_type="text/markdown",
    url=REPO_URL,
    download_url="{repo_url}archive/{version}.tar.gz".format(
        **{"repo_url": REPO_URL, "version": PYPI_VERSION}
    ),
    author="POLITICO interactive news",
    author_email="interactives@politico.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
    ],
    keywords="",
    install_requires=[
        "django",
        "django-uuslug",
        "djangorestframework",
        "boto3",
        "crayons",
        "psycopg2-binary",
        "progress",
    ],
    extras_require={"test": ["pytest"]},
)



# Imports from python.
from itertools import groupby
from io import BytesIO
from operator import itemgetter
import os


# Imports from Django.
from django.apps import apps
from django.core import serializers
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.core.management.utils import parse_apps_and_model_labels
from django.db import DEFAULT_DB_ALIAS
from django.db import connections
# from django.db import router
from django.db import transaction


# Imports from other dependencies.
import crayons
from rest_framework.utils import model_meta


# Imports from civic_utils.
from civic_utils.conf import settings as civic_utils_settings
from civic_utils.management.helpers import ALLOWED_SERIALIZER_FORMATS
from civic_utils.management.helpers import FORMAT_EXTENSION_MAP
from civic_utils.management.helpers import TEMP_DIR
from civic_utils.models import CivicBaseModel
from civic_utils.utils.aws import get_bucket


NATURAL_KEY_SEPARATOR = '___'

S3_FIXTURE_BUCKET = civic_utils_settings.AWS_S3_BUCKET

S3_FIXTURE_ROOT_PATH = civic_utils_settings.FIXTURE_ROOT_PATH


def flatten_natural_key(model_dict, natural_key_fields):
    return NATURAL_KEY_SEPARATOR.join([
        model_dict[field] for field in natural_key_fields
    ])


def get_nk_fieldnames(model_klass, relation_fieldname):
    relation_field = model_klass._meta.get_field(relation_fieldname)

    if relation_field.is_relation and issubclass(model_klass, CivicBaseModel):
        return [
            f"{relation_field.name}__{nk_field}"
            for nk_field
            in relation_field.related_model.get_natural_key_fields()
        ]

    return []


class Command(BaseCommand):
    help = 'Installs the named fixture(s) in the database.'
    help = (
        "Import Civic data fixtures (either from a known S3 bucket location "
        "or a known local directory). "
    )

    def add_arguments(self, parser):
        # parser.add_argument(
        #     'args', metavar='fixture', nargs='+', help='Fixture labels.'
        # )
        parser.add_argument(
            '--database', default=DEFAULT_DB_ALIAS,
            help=(
                "Nominates a specific database to load fixtures into. "
                "Defaults to the 'default' database."
            ),
        )
        # parser.add_argument(
        #     '--ignorenonexistent', '-i', action='store_true', dest='ignore',
        #     help=(
        #         "Ignores entries in the serialized data for fields that "
        #         "do not currently exist on the model."
        #     ),
        # )
        parser.add_argument(
            '-e', '--exclude', action='append', default=[],
            help=(
                "An app_label or app_label.ModelName to exclude. "
                "Can be set multiple times."
            ),
        )
        parser.add_argument(
            '--format', default='json',
            help=(
                "Format of  zed data to be loaded. "
                "Defaults to the format of files in the specified source "
                "directory (if one and only one format exists at that path)."
            ),
        )
        parser.add_argument(
            '-l', '--local',
            action='store_true',
            dest='load_locally',
            help=(
                "Load fixtures from a local directory, rather than from S3."
            ),
        )
        parser.add_argument(
            '-t', '--test',
            action='store_true',
            dest='test_mode',
            help=(
                "Test mode: Only show what would change by loading fixtures "
                "(without making any changes to the database yet)."
            )
        )
        parser.add_argument(
            '-f', '--filename',
            default='export',
            dest='input_filename',
            help=(
                "Specifies filenames from which the fixtures are loaded. "
                "Defaults to 'export.json' (or 'export.xml', or other "
                "extensions based on the specified format). "
                "Useful if you want to store different versions of the same "
                "models' fixtures."
                "Note that this argument only sets the filename without file "
                "extension (as that is automatically configured) and without "
                "directory (as that is a known location, either on your local "
                "system or in an S3 bucket)."
            )
        )

    def handle(self, *fixture_labels, **options):
        # self.ignore = options['ignore']
        self.using = options['database']
        self.verbosity = options['verbosity']
        self.excluded_models, self.excluded_apps = parse_apps_and_model_labels(
            options['exclude']
        )
        self.format = options['format']
        self.load_locally = options['load_locally']
        self.test_mode = options['test_mode']
        self.input_filename = options['input_filename']

        if not self.load_locally:
            self.bucket = get_bucket(S3_FIXTURE_BUCKET)

        # Check that the serialization format exists; this is a shortcut to
        # avoid collating all the objects and _then_ failing.
        if self.format not in ALLOWED_SERIALIZER_FORMATS:
            try:
                serializers.get_serializer(self.format)
            except serializers.SerializerDoesNotExist:
                pass

                raise CommandError("Unknown serialization format: %s" % format)

        self.stdout.write("")
        self.stdout.write(crayons.magenta('Civic Data Importer', bold=True))
        self.stdout.write("")

        self.loadable_models = [
            (app_config, model_klass)
            for app_config in apps.get_app_configs()
            for model_klass in app_config.get_models()
            if issubclass(model_klass, CivicBaseModel)
        ]

        self.fixtures_to_load = self.find_fixtures()

        self.fixture_app_order = self.sort_fixture_imports()

        self.stdout.write('')
        self.stdout.write((
            "The following types and counts of models will be imported "
            f"({len(self.fixtures_to_load)} model types):"
        ))
        self.stdout.write('')

        self.fixtures_by_model = dict(
            (fixture['model_klass'], fixture)
            for fixture in self.fixtures_to_load
        )

        self.fixtures_to_create = {}
        self.fixtures_to_update = {}

        for model_config in self.fixture_app_order:
            (
                manager,
                fixtures_to_create,
                models_to_update,
                fields_to_update,
                post_create_manytomanys,
                post_update_manytomanys,
            ) = self.prepare_model_creation_and_updates(model_config)

            model_verbose_name = model_config._meta.verbose_name

            if not self.test_mode:
                with transaction.atomic(using=self.using):
                    if fixtures_to_create:
                        created_instance_count = len(fixtures_to_create)
                        created_plural = 's' if created_instance_count != 1 else ''
                        creating_message = (
                            f"Saving {created_instance_count} new "
                            f"{model_verbose_name} instance{created_plural}:"
                        )

                        self.stdout.write(
                            crayons.white(creating_message, bold=True),
                            ending='\r'
                        )
                        self.stdout.flush()

                        # Bulk-create new models.
                        manager.bulk_create(fixtures_to_create)

                        self.stdout.write('{} {}'.format(
                            crayons.white(creating_message, bold=True),
                            crayons.green("Done.", bold=True)
                        ))

                    if models_to_update:
                        updated_instance_count = len(models_to_update)
                        updated_plural = 's' if updated_instance_count != 1 else ''
                        updating_message = (
                            f"Saving {updated_instance_count} updated "
                            f"{model_verbose_name} instance{updated_plural}:"
                        )

                        self.stdout.write(
                            crayons.white(updating_message, bold=True),
                            ending='\r'
                        )
                        self.stdout.flush()

                        # Bulk-update existing (but stale) models.
                        manager.bulk_update(models_to_update, fields_to_update)

                        self.stdout.write('{} {}'.format(
                            crayons.white(updating_message, bold=True),
                            crayons.green("Done.", bold=True)
                        ))

                    self.stdout.write("")

            self.fixtures_to_create.setdefault(manager, fixtures_to_create)
            self.fixtures_to_update.setdefault(manager, models_to_update)

        # Close the DB connection -- unless we're still in a transaction. This
        # is required as a workaround for an edge case in MySQL: if the same
        # connection is used to create tables, load data, and query, the query
        # can return incorrect results. See Django #7572, MySQL #37735.
        if transaction.get_autocommit(self.using):
            connections[self.using].close()

    def find_fixtures(self):
        input_extension = FORMAT_EXTENSION_MAP[self.format]
        input_pattern = f"{self.input_filename}.{input_extension}"

        if not self.load_locally:
            all_remote_fixtures = self.bucket.objects.filter(
                Prefix=f'{S3_FIXTURE_ROOT_PATH}/'
            )
            root_path_length = len(S3_FIXTURE_ROOT_PATH)

            # The fixtures we'll attempt to load will be one level below the
            # 'S3_FIXTURE_ROOT_PATH' directory (in folders named for the models
            # they represent).
            # Within these model-specific folders, the fixtures we'll evaluate
            # must have the filename we're looking for -- which is the value of
            # 'input_pattern' as calculated above.
            matching_fixtures = [
                dict(
                    filename=fixture_file.key,
                    model_slug=os.path.split(
                        fixture_file.key
                    )[0][root_path_length:].strip(os.path.sep),
                ) for fixture_file in all_remote_fixtures
                if os.path.split(fixture_file.key)[1] == input_pattern
                and fixture_file.key[:root_path_length] == S3_FIXTURE_ROOT_PATH
            ]
        else:
            temporary_dir_subdirectories = [
                os.path.join(TEMP_DIR, item_path)
                for item_path in os.listdir(TEMP_DIR)
                if os.path.isdir(os.path.join(TEMP_DIR, item_path))
            ]
            matching_fixtures = [
                dict(
                    filename=os.path.join(subdirectory, input_pattern),
                    model_slug=subdirectory[len(TEMP_DIR):].strip(os.path.sep),
                )
                for subdirectory in temporary_dir_subdirectories
                if os.path.isfile(os.path.join(subdirectory, input_pattern))
            ]

        model_slug_map = dict(
            (model_klass._meta.model_name, (app_config, model_klass))
            for (app_config, model_klass) in self.loadable_models
        )

        # Add app and model configurations to fixture context.
        full_context_matching_fixtures = [
            dict(
                app_config=model_slug_map[fixture_config['model_slug']][0],
                model_klass=model_slug_map[fixture_config['model_slug']][1],
                **fixture_config,
            )
            for fixture_config in matching_fixtures
            if fixture_config['model_slug'] in model_slug_map
        ]

        # Filter out fixtures that have been explicitly excluded.
        finalized_matching_fixtures = [
            fixture_config for fixture_config in full_context_matching_fixtures
            if fixture_config['app_config'] not in self.excluded_apps
            and fixture_config['model_klass'] not in self.excluded_models
        ]

        return finalized_matching_fixtures

    def sort_fixture_imports(self):
        app_sorted_fixtures = sorted(
            self.fixtures_to_load,
            key=lambda x: x['app_config'].name
        )

        to_import_list = dict(
            (app_config, [model_cfg['model_klass'] for model_cfg in group])
            for app_config, group in groupby(
                app_sorted_fixtures,
                itemgetter('app_config')
            )
        )

        return serializers.sort_dependencies(to_import_list.items())

    def load_fixture(self, fixture_config):
        if not self.load_locally:
            s3_fileobj = self.bucket.Object(key=fixture_config['filename'])
            with BytesIO(s3_fileobj.get()['Body'].read()) as fixture_input:
                deserialized_data = fixture_config['model_klass'].deserialize(
                    self.format,
                    fixture_input
                )

        with open(fixture_config['filename'], 'rb') as fixture_input:
            deserialized_data = fixture_config['model_klass'].deserialize(
                self.format,
                fixture_input
            )

        return deserialized_data

    def prepare_model_creation_and_updates(self, model_config):
        fixture_config = self.fixtures_by_model[model_config]
        fixture_models = self.load_fixture(fixture_config)

        model_klass = fixture_config['model_klass']

        model_manager = model_klass._default_manager
        model_metadata = model_meta.get_field_info(model_klass)

        nk_fields = model_klass.get_natural_key_fields()

        pk_field = model_klass._meta.pk.name

        compared_fields = ['updated', *nk_fields]

        existing_instance_fields = model_manager.values(
            pk_field,
            *compared_fields
        )
        fixture_model_dicts = [
            {
                '_instance': fixture_instance,
                **fixture_instance.object.__dict__,
                **dict(
                    list(*zip(
                        get_nk_fieldnames(model_klass, relation_fieldname),
                        natural_key_parts
                    ))
                    for relation_fieldname, natural_key_parts
                    in fixture_instance.m2m_data.items()
                ),
                **dict(
                    list(*zip(
                        get_nk_fieldnames(model_klass, relation_fieldname),
                        natural_key_parts
                    ))
                    for relation_fieldname, natural_key_parts
                    in fixture_instance.deferred_fields.items()
                ),
            }
            for fixture_instance in fixture_models
        ]
        fixture_instance_fields = [
            {
                '_instance': fixture_dict['_instance'],
                **{
                    field_name: fixture_dict[field_name]
                    for field_name in compared_fields
                }
            }
            for fixture_dict in fixture_model_dicts
        ]

        existing_summary = {
            flatten_natural_key(instance, nk_fields): instance['updated']
            for instance in existing_instance_fields
        }
        existing_natural_key_id_map = {
            flatten_natural_key(instance, nk_fields): instance[pk_field]
            for instance in existing_instance_fields
        }
        fixture_summary = {
            flatten_natural_key(instance, nk_fields): instance['updated']
            for instance in fixture_instance_fields
        }
        fixture_model_instances = {
            flatten_natural_key(instance, nk_fields): instance['_instance']
            for instance in fixture_instance_fields
        }

        only_existing = existing_summary.keys() - fixture_summary.keys()
        only_fixture = fixture_summary.keys() - existing_summary.keys()
        overlapping_instances = set(
            fixture_summary.keys()
        ).intersection(existing_summary.keys())

        existing_newer = []
        fixture_newer = []
        equal_age = []

        for natural_key in overlapping_instances:
            existing_updated_date = existing_summary[natural_key]
            fixture_updated_date = fixture_summary[natural_key]

            if existing_updated_date > fixture_updated_date:
                existing_newer.append(natural_key)
            elif existing_updated_date < fixture_updated_date:
                fixture_newer.append(natural_key)
            else:
                equal_age.append(natural_key)


        self.stdout.write('')
        self.stdout.write(crayons.yellow(
            f"[{model_klass._meta.label}]",
            bold=True
        ))
        self.stdout.write('')
        self.stdout.write(f"-   Only in database: {len(only_existing)}")
        self.stdout.write(
            f"-   In both; database newer: {len(existing_newer)}"
        )
        self.stdout.write(f"-   In both; equal age: {len(equal_age)}")
        self.stdout.write(crayons.cyan(
            f"-   * In both; fixture newer: {len(fixture_newer)}", bold=True
        ))
        self.stdout.write(crayons.cyan(
            f"-   * Only in fixture: {len(only_fixture)}", bold=True
        ))
        self.stdout.write('')

        total_modifications = len(fixture_newer) + len(only_fixture)
        if total_modifications > 0:
            self.stdout.write(crayons.cyan(
                f"    * Total to be saved: {total_modifications}", bold=True
            ))
            self.stdout.write('')

        current_related_model_keys = {}
        for related_field_name in model_metadata.relations.keys():
            related_field = model_klass._meta.get_field(related_field_name)
            related_model = related_field.related_model

            raw_related_values = related_model._default_manager.values(
                related_model._meta.pk.name,
                *related_model.get_natural_key_fields()
            )

            current_related_model_keys[related_field.name] = dict((
                flatten_natural_key(
                    related_instance,
                    related_model.get_natural_key_fields()
                ),
                related_instance[related_model._meta.pk.name]
            )  for related_instance in raw_related_values)

        fixtures_to_be_created = []
        models_to_be_updated = []
        fields_used_in_update = set()

        if fixture_newer:
            models_needing_update = model_manager.filter(
                **{f"{pk_field}__in": [
                    existing_natural_key_id_map[natural_key]
                    for natural_key in fixture_newer
                ]}
            )
        else:
            models_needing_update = model_manager.none()

        existing_id_instance_map = {
            getattr(instance, pk_field): instance
            for instance in models_needing_update
        }

        # Calculate which fields we can and should update on this model.
        allowed_update_fields = [
            field_config.name
            for field_config in model_klass._meta.get_fields()
            # 1. No auto-created (reverse-direction relational) fields.
            if not field_config.auto_created
            # 2. No primary keys.
            and not field_config.primary_key
            # 3. No fields used to calculate PKs.
            and field_config.name
            not in model_klass.get_natural_key_definition()
        ]

        for instance_key, instance_data in fixture_model_instances.items():
            if instance_key in only_fixture:
                evaluated = self.evaluate_deferred(
                    model_klass,
                    current_related_model_keys,
                    instance_data
                )
                fixtures_to_be_created.append(evaluated)
            elif instance_key in fixture_newer:
                updated_instance = self.evaluate_deferred(
                    model_klass,
                    current_related_model_keys,
                    instance_data
                )
                current_pk = existing_natural_key_id_map[instance_key]
                current_instance = existing_id_instance_map[current_pk]

                apply_instance_update = False

                for field_to_update in allowed_update_fields:
                    current_value = getattr(current_instance, field_to_update)
                    updated_value = getattr(updated_instance, field_to_update)

                    if current_value != updated_value:
                        apply_instance_update = True
                        fields_used_in_update.add(field_to_update)
                        setattr(
                            current_instance,
                            field_to_update,
                            updated_value
                        )

                if apply_instance_update:
                    models_to_be_updated.append(current_instance)

        return (
            model_manager,
            fixtures_to_be_created,
            models_to_be_updated,
            list(fields_used_in_update),
            [],  # Post-create M2Ms.
            [],  # Post-edit M2Ms.
        )

    def evaluate_deferred(self, model_klass, related_id_maps, instance):
        for deferred_field, natural_key in instance.deferred_fields.items():
            related_field = model_klass._meta.get_field(deferred_field)
            rel_nk_parts = related_field.related_model.get_natural_key_fields()
            rel_nk_values = dict(zip(rel_nk_parts, natural_key))

            flat_natural_key = flatten_natural_key(rel_nk_values, rel_nk_parts)

            if deferred_field not in related_id_maps:
                raise ValueError(
                    "Could not find natural-key to ID mapping for "
                    f"'{deferred_field}' field."
                )
            elif flat_natural_key not in related_id_maps[deferred_field]:
                self.stdout.write('{}{}'.format(
                    crayons.red('WARNING: ', bold=True),
                    (
                        "No match was found for field "
                        f"'{deferred_field}' and natural key "
                    )
                ))
                self.stdout.write(f"         '{natural_key}' when creating or modifying ")
                self.stdout.write(
                    f"         a(n) '{model_klass._meta.label}' "
                    "model instance."
                )
                self.stdout.write('')
                self.stdout.write("         Records of this relationship may be lost.")
                self.stdout.write('')
                continue

            pk_for_relation = related_id_maps[deferred_field][flat_natural_key]

            setattr(instance.object, f"{deferred_field}_id", pk_for_relation)

        return instance.object



# Imports from python.
from collections import OrderedDict
from io import StringIO


# Imports from Django.
from django.core.serializers.base import DeserializedObject
from django.core.serializers.base import ProgressBar
from django.db import models
from django.utils.functional import cached_property


# Imports from other dependencies.
from rest_framework.exceptions import ValidationError
from rest_framework import serializers
from rest_framework.renderers import JSONRenderer
from rest_framework.utils import model_meta
from rest_framework.validators import UniqueValidator


# Imports from civic_utils.
from civic_utils.models import CivicBaseModel


class CommandLineSerializerMixin(object):
    def serialize_to_cli(
        self,
        data=None,
        stream=None,
        progress_output=None,
        object_count=0,
        renderer=None,
        **options,
    ):
        if data is None:
            data = self.instance

        self.cli_options = options
        self.cli_stream = stream if stream is not None else StringIO()
        progress_bar = ProgressBar(progress_output, object_count)

        self.start_cli_serialization()
        self.cli_first_item = True
        self.cli_renderer_klass = renderer or JSONRenderer

        iterable = data.all() if isinstance(data, models.Manager) else data

        for count, obj in enumerate(iterable, start=1):
            self.start_object(obj)
            self.end_object(obj)

            progress_bar.update(count)

            self.cli_first_item = self.cli_first_item and False

        self.end_cli_serialization()
        return self.getvalue()

    def start_cli_serialization(self):
        self._current = None
        self.cli_exported_objects = []

    def end_cli_serialization(self):
        renderer_context = {}

        if "indent" in self.cli_options:
            renderer_context["indent"] = self.cli_options["indent"]

        self.cli_stream.write(
            self.cli_renderer_klass()
            .render(
                self.cli_exported_objects, renderer_context=renderer_context
            )
            .decode("utf-8")
        )

    def start_object(self, obj):
        self._current = self.child.to_representation(obj)

    def end_object(self, obj):
        self.cli_exported_objects.append(self._current)
        self._current = None

    def getvalue(self):
        """
        Return the fully serialized queryset (or None if the output stream is
        not seekable).
        """
        if callable(getattr(self.cli_stream, "getvalue", None)):
            return self.cli_stream.getvalue()


class CommandLineDeserializerMixin(object):
    def to_python(self, raw_data):
        if hasattr(self, "many") and self.many:
            return [self.child.to_python(instance) for instance in raw_data]

        model_klass = self.Meta.model

        info = model_meta.get_field_info(model_klass)

        m2m_naturalkeys = {}
        deferred_fields = {}

        for field_name, relation_info in info.relations.items():
            if relation_info.reverse is True:
                continue

            if not issubclass(relation_info.related_model, CivicBaseModel):
                continue

            relationship_name = relation_info.model_field.name

            related_nk_fields = relation_info.related_model.natural_key_fields

            relation_natural_key_fields = [
                f"{relationship_name}__{natural_key_component}"
                for natural_key_component in related_nk_fields
            ]

            if not all(
                [field in raw_data for field in relation_natural_key_fields]
            ):
                raise ValueError(
                    "Missing one or more fields needed to construct the "
                    f"natural key for '{relationship_name}'."
                )

            current_natural_key = [
                raw_data.pop(field) for field in relation_natural_key_fields
            ]

            if relation_info.to_many:
                m2m_naturalkeys[relationship_name] = current_natural_key
            else:
                deferred_fields[relationship_name] = current_natural_key

        return DeserializedObject(
            model_klass(**raw_data), m2m_naturalkeys, deferred_fields
        )

    def deserialize(self, initial_data):
        self.initial_data = initial_data

        if not self.is_valid():
            raise ValueError(self.errors)

        return self.to_python(self.validated_data)


class CommandLineCapableListSerializer(
    CommandLineSerializerMixin,
    CommandLineDeserializerMixin,
    serializers.ListSerializer,
):
    pass


class CommandLineListSerializer(
    CommandLineDeserializerMixin, serializers.ModelSerializer
):
    created = serializers.DateTimeField(format="iso-8601")
    updated = serializers.DateTimeField(format="iso-8601")

    class Meta:
        list_serializer_class = CommandLineCapableListSerializer


class NaturalKeyValidator(serializers.UniqueTogetherValidator):
    def set_context(self, serializer):
        self.nested_fields = {
            name: serializer.fields[name]
            for name in self.fields
            if isinstance(serializer.fields[name], NaturalKeySerializer)
        }
        super(NaturalKeyValidator, self).set_context(serializer)

    def filter_queryset(self, attrs, queryset):
        attrs = attrs.copy()
        for field in attrs:
            if field in self.nested_fields:
                assert isinstance(attrs[field], dict)
                cls = self.nested_fields[field].Meta.model
                result = cls._default_manager.filter(**attrs[field])
                if result.count() == 0:
                    # No existing nested object for these values
                    return queryset.none()
                else:
                    # Existing nested object, use it to validate
                    attrs[field] = result[0].pk

        return super(NaturalKeyValidator, self).filter_queryset(
            attrs, queryset
        )


class NaturalKeySerializer(serializers.ModelSerializer):
    """Self-nesting serializer for natural-keyed models."""

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super(
            NaturalKeySerializer, self
        ).build_standard_field(field_name, model_field)

        if "validators" in field_kwargs:
            field_kwargs["validators"] = [
                validator
                for validator in field_kwargs.get("validators", [])
                if not isinstance(validator, UniqueValidator)
            ]
        return field_class, field_kwargs

    def build_nested_field(self, field_name, relation_info, nested_depth):
        field_class = NaturalKeySerializer.for_model(
            relation_info.related_model, validate_key=False
        )
        field_kwargs = {}
        return field_class, field_kwargs

    # def create(self, validated_data):
    #     model_class = self.Meta.model
    #     natural_key_fields = model_class.get_natural_key_fields()
    #     natural_key = []
    #     for field in natural_key_fields:
    #         val = validated_data
    #         for key in field.split('__'):
    #             val = val[key]
    #         natural_key.append(val)
    #     return model_class.objects.find(*natural_key)
    #
    # def update(self, instance, validated_data):
    #     raise NotImplementedError(
    #         "Updating an existing natural key is not supported."
    #     )

    @classmethod
    def for_model(cls, model_class, validate_key=True, include_fields=None):
        unique_together = model_class.get_natural_key_definition()
        if include_fields and list(include_fields) != list(unique_together):
            raise NotImplementedError(
                "NaturalKeySerializer for '%s' has unique_together = [%s], "
                "but provided include_fields = [%s]"
                % (
                    model_class._meta.model_name,
                    ", ".join(unique_together),
                    ", ".join(include_fields),
                )
            )

        class Serializer(cls):
            class Meta(cls.Meta):
                model = model_class
                fields = unique_together
                if validate_key:
                    validators = [
                        NaturalKeyValidator(
                            queryset=model_class._default_manager,
                            fields=unique_together,
                        )
                    ]
                else:
                    validators = []

        return Serializer

    @classmethod
    def for_depth(cls, model_class):
        return cls

    class Meta:
        depth = 1


class NaturalKeySerializerMixin(object):
    """Natural key support across related models.

    Properly serializes models that have one or more foreign keys to
    natural-keyed model(s).
    """

    pk = serializers.SerializerMethodField(method_name="get_natural_key")

    @cached_property
    def fields(self):
        raw_fields = super(NaturalKeySerializerMixin, self).fields

        # Add the 'pk' field to output.
        raw_fields["pk"] = self.pk

        # Make 'pk' the first field displayed.
        raw_fields.fields.move_to_end("pk", last=False)

        return raw_fields

    def get_natural_key(self, obj):
        return obj.natural_key()

    def build_nested_field(self, field_name, relation_info, nested_depth):
        if issubclass(relation_info.related_model, CivicBaseModel):
            field_class = NaturalKeySerializer.for_model(
                relation_info.related_model, validate_key=False
            )

            field_kwargs = {}
            if relation_info.model_field is not None:
                if relation_info.model_field.null:
                    field_kwargs["required"] = False
            return field_class, field_kwargs

        return super(NaturalKeySerializerMixin, self).build_nested_field(
            field_name, relation_info, nested_depth
        )

    def build_relational_field(self, field_name, relation_info):
        field_class, field_kwargs = super(
            NaturalKeySerializerMixin, self
        ).build_relational_field(field_name, relation_info)
        if issubclass(relation_info.related_model, CivicBaseModel):
            field_kwargs.pop("queryset")
            field_kwargs["read_only"] = True
        return field_class, field_kwargs

    def get_fields(self):
        fields = super(NaturalKeySerializerMixin, self).get_fields()
        fields.update(self.build_natural_key_fields())
        return fields

    def build_natural_key_fields(self):
        info = model_meta.get_field_info(self.Meta.model)
        fields = OrderedDict()
        for field, relation_info in info.relations.items():
            if not issubclass(relation_info.related_model, CivicBaseModel):
                continue

            if field in self.fields:
                field_class, field_kwargs = self.build_nested_field(
                    field, relation_info, 1
                )
                fields[field] = field_class(**field_kwargs)
        return fields

    # def create(self, validated_data):
    #     self.convert_natural_keys(
    #         validated_data
    #     )
    #     return super(NaturalKeySerializerMixin, self).create(
    #         validated_data
    #     )
    #
    # def update(self, instance, validated_data):
    #     self.convert_natural_keys(
    #         validated_data
    #     )
    #     return super(NaturalKeySerializerMixin, self).update(
    #         instance, validated_data
    #     )
    #
    # def convert_natural_keys(self, validated_data):
    #     fields = self.get_fields()
    #     for name, field in fields.items():
    #         if name not in validated_data:
    #             continue
    #         if isinstance(field, NaturalKeySerializer):
    #             validated_data[name] = fields[name].create(
    #                 validated_data[name]
    #             )

    @classmethod
    def for_model(cls, model_class, include_fields=None):
        # c.f. wq.db.rest.serializers.ModelSerializer
        class Serializer(cls):
            class Meta(cls.Meta):
                model = model_class
                if include_fields:
                    fields = include_fields

        return Serializer

    def get_unique_constraint_fields(self):
        return [
            field.name
            for field in self.Meta.model._meta.get_fields()
            if hasattr(field, "unique") and field.unique
        ]

    def run_validation(self, data):
        raw_natural_key_fields = data.pop("pk")

        cached_unique_fields = {}
        unique_fields = self.get_unique_constraint_fields()
        for field in unique_fields:
            if field in data:
                cached_unique_fields[field] = data[field]

        natural_key_components = dict(
            zip(
                self.Meta.model.get_natural_key_fields(),
                raw_natural_key_fields,
            )
        )

        validated = super(NaturalKeySerializerMixin, self).run_validation(data)

        validated.update(natural_key_components)

        for field_name, cached_value in cached_unique_fields.items():
            if field_name not in validated:
                validated.update(dict([(field_name, cached_value)]))

        return validated

    class Meta:
        pass
